/*  loop.c
    Copyright (C) 1993  Grant Philip Maizels

    $Header: /usr/src/dp/RCS/loop.c,v 0.6 1995/03/31 10:05:48 gpm Exp $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* These routines do a pre scan of the program to detect forward references
 * to blocks.  This means we can put labels on only thos blocks which
 * need them.
 */

#include "dp.h"

char loops[MAX_LOOPS];

void l_unchain(us start)
{
    us a, c, n;

    a = start;
    while(1)
    {
	c = gs(a + 2);
	n = gs(a);
	l_discmd(a, c);
	if (n == 0) break;
	a = n;
    }
}

void l_discmd(us a, us c)
{
    switch(c)
    {
	case 0x64:	l_if(a);        break;
	case 0xc8:	l_block(a);     break;
	case 0x259:	l_loopctl(a);   break;
	case 0x2c9:	l_assign(a);    break;
    }
}

void l_assign(us a)
{
    if (gs(a+10) != 0 && (gb(a+9) & 2) == 2)
    {
	l_disblk(gs(a+10), 4);
    }
}

void l_loopctl(us a)
{
    if (gs(a + 4) != 0xffff)
    {
	loops[gs(a + 4)] |= L_USED;
	if (gs(a+4) != curr_blk)
	    loops[gs(a + 4)] |= L_NAMED;
    }
    if (gs(a+8) == 0xe880 && gs(a+6) == 0)
	loops[0] |= L_USED;
    else
    {
	loops[gs(a + 6)] |= L_USED;
	if (gs(a+6) != curr_blk)
	    loops[gs(a + 6)] |= L_NAMED;
    }
}
    
void l_block(us a)
{
    l_disblk(gs(a+4), -1);
}

void l_if(us a)
{
    if(gs(a + 6) != 0)
	l_unchain(gs(a + 6));
    if(gs(a + 8) != 0)
	l_unchain(gs(a + 8));
}

void l_disblk(us a, int lt)
{
    int ltype; /* 0 = procedure, 1 = do, 2 = repeat, 3 = for, 4 = editing,
		  -1 dont't know */
    int save_blk;

    ltype = lt;
    if (ltype == -1)
    {
	if (gs(a+6) == 0 && gs(a+16) == 0 && gs(a+18) == 0)
	    ltype = 1;
	else
	    if (gs(a+6) == 0)
		ltype = 2;
	    else
		ltype = 3;
    }
    loops[gs(a)] |= ltype;
    save_blk = curr_blk;
    curr_blk = gs(a);
    if (gs(a+14) != 0)
	l_unchain(gs(a + 14));
    if (gs(a+20) != 0)
	l_unchain(gs(a + 20));
    if (gs(a+16) != 0 &&
	(ltype == 1 ||
	 gs(gs(a+16) + 4) != curr_blk ||
	 gs(gs(a+16) + 6) != curr_blk ||
	 gs(gs(a+16) + 8) != 0xf752))
	l_unchain(gs(a + 16));
    if (gs(a+18) != 0 &&
	(ltype == 1 ||
	 gs(gs(a+18) + 4) != curr_blk ||
	 gs(gs(a+18) + 6) != curr_blk ||
	 gs(gs(a+18) + 8) != 0xe865))
	l_unchain(gs(a + 18));
    l_unchain(gs(a + 10));
    curr_blk = save_blk;
}
