/*  misc.c
    Copyright (C) 1993  Grant Philip Maizels

    $Header: /usr/src/dp/RCS/misc.c,v 0.6 1995/03/31 10:05:49 gpm Exp $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "dp.h"
#include <unistd.h>
/* #include <varargs.h> */
#include <stdarg.h>

int nl_indent_pending = 0;

unsigned long gl(int i)
{
  if (byteswapped)
    return (((unsigned long) p[i+3] * 256 * 65536) + 
	   ((unsigned long) p[i+2] * 65536) + 
	   ((unsigned long) p[i+1] * 256) + 
	   (unsigned long) p[i]);
  else
    return (((unsigned long) p[i] * 256 * 65536) +
	   ((unsigned long) p[i+1] * 65536) + 
	   ((unsigned long) p[i+2] * 256) + 
	   (unsigned long) p[i+3]);
}

us gs(us i)
{
    if (byteswapped)
        return ((p[i+1] * 256) + p[i]);
    else
	return ((p[i] * 256) + p[i+1]);
}

unsigned char gb(int i)
{
    return (p[i]);
}

char *indent()
{
    static char s[132];
    int i, lev;

    lev = ind_lev * 4;
    if (lev > 80)
	lev = 80;
    for (i = 0; i < 132; i++)
       s[i] = ' ';
    s[lev] = '\0';
    return(s);
}

void usage()
{
    fprintf(stderr, 
	"dp [-NPVi] [-d debugcode] [-h hexdumps] [-v version] file\n");
    exit(2);
}

char *vtype(us vt)
{
    switch(vt)
    {
	case 0:      return("unknown");
	case T_CHA:  return("character");
	case T_DAT:  return("date");
	case T_LOG:  return("logical");
	case T_INT:  return("integer");
	case T_DEC:  return("decimal");
	case T_RID:  return("recid");
	case T_RAW:  return("raw");
	default:    return(mpf("(?unknown %04x)", vt));
    }
}

void faterror(char *fmt, ...) /* fatal error */
{
    va_list args;

    va_start(args, fmt);
    (void)fprintf(ignfaterr?stdout:stderr, "Fatal Error:");
    (void)vfprintf(ignfaterr?stdout:stderr, fmt, args);
    va_end(args);
/*     (void)abort(); */
    if (! ignfaterr)
	exit(1);
}

char *newe(char *p, char *msg)
{
    int i;
    char *ptr, msgbuf[50];

    i = strlen(p);
    ptr = malloc(i + 2);
    if (ptr == NULL)
	faterror("No memory for %s", msg);
    strcpy(ptr, p);
    return(ptr);
}

char *mpf(char *fmt, ...) /* sprintf into malloced space */
{
    va_list args;
    char buffer[24576];

    va_start(args, fmt);
    (void)vsprintf(buffer, fmt, args);
    va_end(args);
    return(newe(buffer, "mpf"));
}

void propf(char *fmt, ...) /* like printf() for decompiled output */
{
    va_list args;
    char *nbuf;
    char buf[24576];
    int i, j, k, cnt, ind;

    if (pass == 1)
	return;
    ind = ind_lev;
    if (ind < 0) 
	ind = 0;
    if(nl_indent_pending)
	for(i=0; i < ind; i++)
	    printf("    ");
    nl_indent_pending = 0;
    va_start(args, fmt);
    (void)vsprintf(buf, fmt, args);
    cnt=0;
    for (i=0; i < strlen(buf); i++)
	if(buf[i] == '\n')
	    cnt++;
    nbuf = malloc(strlen(buf) + (cnt * ind * 4) + 2);
    if (nbuf == NULL)
    {
	fprintf(stderr, "No space\n");
	exit(1);
    }
    j = 0;
    for (i=0; i < strlen(buf) + 1; i++)
    {
	nbuf[j++] = buf[i];
	if(buf[i] == '\n')
	    for(k=0; k<4*ind; k++)
		if(buf[i+1] == '\0')
		    nl_indent_pending = 1;
		else
		    nbuf[j++] = ' ';
    }
    printf("%s", nbuf);
    fflush(stdout);
    free(nbuf);
    va_end(args);
}

void dbpf(char dbf, char *fmt, ...) /* like printf() for debugging output */
{
    va_list args;

    if (pass == 1 && dbp1 == 0)
	return;
    va_start(args, fmt);
    if((dbf == 'V' && Vdebug) ||
       (dbf == 'l' && lflag) ||
       (dbf == 'v' && vdebug) ||
       (dbf == 'f' && fdebug) ||
       (dbf == 's' && sdebug) ||
       (dbf == 'e' && edebug) ||
       (dbf == 'a' && adebug))
	(void)vfprintf(stdout, fmt, args);
    fflush(stdout);
    va_end(args);
}

char *fndpth(char *pathlst, char *file)
{
    int cnt = 0, i, len;
    char buf[1024];
    char *npath, *pathe;

    len = strlen(pathlst);
    npath = malloc(len + 1);
    if (npath == NULL)
    {
	 fprintf(stderr, "No space\n");
	 exit(1);
    }
    strncpy(npath, pathlst, len + 1);
    for (i = 0; i < len; i++)
	if (pathlst[i] == ':')
	{
	    cnt++;
	    npath[i] = '\0';
	}
    pathe = npath;
    for (i = 0; i <= cnt; i++)
    {
	sprintf(buf, "%s%s%s", pathe, pathe[0] == '\0'?"":"/", file);
	if(access(buf, R_OK) == 0)
	{
	    free(npath);
	    npath = malloc(strlen(buf) + 1);
	    if (npath == NULL)
	    {
		 fprintf(stderr, "No space\n");
		 exit(1);
	    }
	    strncpy(npath, buf, strlen(buf) + 1);
	    return(npath);
	}
	pathe = pathe + strlen(pathe) + 1;
    }
    free(npath);
    return(NULL);
}
